import base64
import importlib
import inspect
import io
import json
import os.path
import types
import typing
import zipfile
from fnmatch import fnmatch

import boto3
import troposphere
import troposphere.awslambda

from lovage.backends import base
from lovage.backends.awslambda import cf
from lovage.backends.base import Serializer
from lovage.dirtools import Dir
from lovage.exceptions import LovageRemoteException, LovageDeploymentException, LovageInternalException
from lovage.utils import is_in_cloud


class ConsistentZipFile(zipfile.ZipFile):
    def add_file(self, local_path, zip_path):
        info = zipfile.ZipInfo(zip_path)
        # set permissions for windows machines so we don't get permission denied on Lambda
        info.external_attr = 0o755 << 16
        # force constant timestamp so same code produces same zip
        info.date_time = (2020, 1, 1, 0, 0, 0)
        self.writestr(info, open(local_path, "rb").read(), zipfile.ZIP_DEFLATED)


class AwsLambdaBackend(base.Backend):
    def __init__(self, instance_name: str, profile_name: str = None):
        self._instance_name = instance_name
        self._functions = []
        if profile_name and not is_in_cloud():
            self._session = boto3.Session(profile_name=profile_name)
        else:
            self._session = boto3.Session()
        self._executor = AwsLambdaExecutor(instance_name, self._session)
        self._additional_resources: typing.List[troposphere.BaseAWSObject] = []
        self._env: typing.Dict[str, object] = {"LOVAGE_IN_CLOUD": "1"}
        self._policies = []
        self._exception_handler = _empty_exception_handler

    def new_task(self, serializer: base.Serializer, func: types.FunctionType, options: typing.Mapping) -> base.Task:
        desc = {
            "Name": _func_lambda_name(func, self._instance_name),
            "CfName": _func_cf_name(func),
            "Handler": _function_lambda_spec(func),
            "Policies": options.get("aws_policies", []),
            "Kwargs": {},
            "OriginalFunction": func,
        }
        if "timeout" in options:
            desc["Kwargs"]["Timeout"] = options["timeout"]
        if "aws_vpc_subnet_ids" in options and "aws_vpc_security_group_ids" in options:
            desc["Kwargs"]["VpcConfig"] = troposphere.awslambda.VPCConfig(
                SubnetIds=options["aws_vpc_subnet_ids"],
                SecurityGroupIds=options["aws_vpc_security_group_ids"],
            )
            desc["Policies"].append({
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Effect": "Allow",
                        "Action": [
                            "ec2:CreateNetworkInterface"
                        ],
                        "Resource": "*"
                    },
                    {
                        "Effect": "Allow",
                        "Action": [
                            "ec2:DescribeNetworkInterfaces",
                            "ec2:DeleteNetworkInterface"
                        ],
                        "Resource": "*"
                        # troposphere.Sub("arn:${AWS::Partition}:ec2:${AWS::Region}:${AWS::AccountId}:network-interface/*")
                    }
                ]
            })
        elif "aws_vpc_security_group_ids" in options or "aws_vpc_subnet_ids" in options:
            raise ValueError("aws_vpc_security_group_ids and aws_vpc_security_group_ids must be used together")
        self._functions.append(desc)
        return AwsTask(func, self._executor, serializer, self._exception_handler)

    def deploy(self, *, requirements: typing.List[str], root: str, exclude=None):
        # TODO allow configuration of this
        # all files in CWD
        # all files in certain directory
        # all python files
        # git archive
        # .gitignore?
        # serverless way
        zs = io.BytesIO()
        packaged_modules = set()
        with ConsistentZipFile(zs, "w") as z:
            exclude = exclude or []
            for walk_root, folders, files in Dir(directory=root, excludes=exclude, exclude_file=".lovageignore").walk():
                for f in files:
                    local_path = os.path.join(walk_root, f)
                    zip_path = os.path.relpath(local_path, '.')
                    z.add_file(local_path, zip_path)
                    packaged_modules.add(os.path.abspath(local_path))

            from lovage import __version__ as lovage_version
            if lovage_version != "0.0.0":
                # prepend requirements to allow user to override
                requirements.insert(0, f"lovage=={lovage_version}")
            else:
                print("Unable to find Lovage version, using local files (can happen while developing Lovage)")
                # lovage dependencies
                requirements.insert(0, "troposphere")
                requirements.insert(1, "globster==0.1.0")

                import lovage
                lovage_dir = os.path.dirname(os.path.dirname(lovage.__file__))

                for walk_root, folders, files in os.walk(lovage_dir):
                    for f in files:
                        local_path = os.path.join(walk_root, f)
                        zip_path = os.path.relpath(local_path, lovage_dir)
                        if fnmatch(zip_path, "lovage/*.py"):
                            z.add_file(local_path, zip_path)

        zs.seek(0)
        zip_bytes = zs.read()

        missing_files = False
        for fd in self._functions:
            f = fd["OriginalFunction"]
            fm = os.path.abspath(inspect.getfile(f))
            if fm not in packaged_modules:
                print(f"{inspect.getmodule(f).__name__}.{f.__name__} is defined in {fm} but it was not packaged")
                missing_files = True

        if missing_files:
            raise LovageDeploymentException(f"Some files are missing from the packaged code, "
                                            f"is root='{root}' the correct setting?")

        cf.deploy(self._session, self._instance_name, zip_bytes, requirements,
                  self._functions, self._additional_resources, self._env, self._policies)

    def add_resource(self, resource: troposphere.BaseAWSObject):
        # TODO better name than resource since this can be output too?
        self._additional_resources.append(resource)

    def add_environment_variable(self, name: str, value):
        # TODO do something that works for local, aws, gcp, etc.
        self._env[name] = value

    def add_common_policy(self, policy):
        self._policies.append(policy)

    def set_exception_handler(self, handler: typing.Callable[[Exception], None]):
        self._exception_handler = handler

    def function_arn(self, func: types.FunctionType):
        return troposphere.GetAtt(_func_cf_name(func), "Arn")


class AwsLambdaExecutor(base.Executor):
    def __init__(self, instance_name: str, session: boto3.Session):
        self._lambda = session.client("lambda")
        self._name = instance_name

    def invoke(self, serializer: base.Serializer, func: types.FunctionType, packed_args):
        result = self._invoke(func, packed_args, "RequestResponse", 200)
        function_result = json.loads(result["Payload"].read())
        if "exception" in function_result:
            # TODO serialize stack trace
            # exceptions coming from here are not really from here, they're from the Lambda function
            exception_data = serializer.unpack_result(base64.b85decode(function_result["exception"]))
            if serializer.objects_supported:
                raise exception_data  # exception from the Lambda function
            else:
                raise LovageRemoteException.from_exception_object(exception_data)  # exception from the Lambda function
        return base64.b85decode(function_result["result"])

    def invoke_async(self, serializer: base.Serializer, func: types.FunctionType, packed_args):
        self._invoke(func, packed_args, "Event", 202)

    def _invoke(self, func: types.FunctionType, packed_args, invocation_type: str, required_status_code: int):
        result = self._lambda.invoke(
            FunctionName=_func_lambda_name(func, self._name),
            InvocationType=invocation_type,
            Payload=json.dumps({
                "packed_args": base64.b85encode(packed_args).decode("utf-8"),
            }),
        )
        if result["StatusCode"] != required_status_code or result.get("FunctionError"):
            error = json.loads(result["Payload"].read())["errorMessage"]
            raise LovageInternalException(f"Unhandled Lambda error for {func.__module__}.{func.__name__}: {error}")

        return result

    def queue(self, serializer: base.Serializer, func: types.FunctionType, packed_args):
        raise NotImplementedError()

    def delay(self, serializer: base.Serializer, func: types.FunctionType, packed_args, timeout):
        raise NotImplementedError()


class AwsTask(base.Task):
    def __init__(self, func: types.FunctionType, executor: AwsLambdaExecutor, serializer: Serializer,
                 exception_handler: typing.Callable[[Exception], None]):
        super().__init__(func, executor, serializer)
        self._exception_handler = exception_handler

    def __call__(self, *args, **kwargs):
        if not is_in_cloud():
            super().call(*args, **kwargs)
        else:
            # this assumes we have the same backend settings here as we do when this was deployed.
            # we use to just tell lambda which serializer, exception handler and function to dynamically load, but this
            # solution was not secure. users can force lambda to execute arbitrary code this way.
            # TODO verify same backend settings with a hash or something?
            event, context = args
            args, kwargs = self._serializer.unpack_args(base64.b85decode(event["packed_args"]))
            try:
                result = self._func(*args, **kwargs)
            except Exception as e:
                self._exception_handler(e)
                if self._serializer.objects_supported:
                    packed_e = self._serializer.pack_result(e)
                else:
                    packed_e = self._serializer.pack_result(LovageRemoteException.exception_object(e))
                return {"exception": base64.b85encode(packed_e).decode("utf-8")}
            packed_result = self._serializer.pack_result(result)
            return {"result": base64.b85encode(packed_result).decode("utf-8")}


def _func_lambda_name(func: types.FunctionType, instance_name) -> str:
    return f"{instance_name}-{_function_spec(func).replace('.', '-').replace(':', '--')}"


def _func_cf_name(func: types.FunctionType) -> str:
    return _function_spec(func).replace(".", "XdotX").replace(":", "XcolonX").replace("_", "XusX")


def _function_lambda_spec(func: types.FunctionType) -> str:
    return _function_spec(func).replace(":", ".")


def _function_spec(func: types.FunctionType) -> str:
    if func.__module__ == "__main__":
        import __main__
        path = __main__.__file__
        relative = os.path.relpath(path, os.getcwd())  # TODO something better than cwd
        module_path = ".".join(os.path.split(os.path.splitext(relative)[0])).strip(".")
        return f"{module_path}:{func.__name__}"
    return f"{func.__module__}:{func.__name__}"


def _empty_exception_handler(e):
    pass




import os
import traceback

import boto3
import cfnresponse


def handler(event, context):
    key = "BAD-PARAMETERS"

    try:
        key = event["ResourceProperties"]["Key"]
        bucket = os.environ['BUCKET']

        if event["RequestType"] in ["Create", "Update"]:
            print(f"Nothing to do for Create or Update")

        elif event["RequestType"] == "Delete":
            print(f"Deleting s3://{bucket}/{key}")

            try:
                boto3.client("s3").delete_object(Bucket=bucket, Key=key)
            except botocore.exceptions.ClientError as e:
                print(f"Error deleting {key}")
                try:
                    traceback.print_last()
                except ValueError:
                    print("Caught exception but unable to print stack trace")
                    print(e)

        result = {"Bucket": bucket, "Key": key}
        cfnresponse.send(event, context, cfnresponse.SUCCESS, result, key)
    except Exception as e:
        try:
            traceback.print_last()
        except ValueError:
            print("Caught exception but unable to print stack trace")
            print(e)
        cfnresponse.send(event, context, cfnresponse.FAILED, {}, key)



import json
import pickle
import types
import typing
import warnings

from lovage.exceptions import LovageException, LovageConfigurationError
from lovage.utils import is_in_cloud

class Serializer(object):
    def __init__(self):
        self.objects_supported = False

    def pack_args(self, args, kwargs):
        return self._serialize({"args": args, "kwargs": kwargs})

    def unpack_args(self, packed_args):
        args = self._deserialize(packed_args)
        return args["args"], args["kwargs"]

    def pack_result(self, result):
        return self._serialize(result)

    def unpack_result(self, packed_result):
        return self._deserialize(packed_result)

    def _serialize(self, obj: typing.Any) -> bytes:
        raise NotImplementedError()

    def _deserialize(self, data: bytes) -> typing.Any:
        raise NotImplementedError()


class PickleSerializer(Serializer):
    def __init__(self):
        super().__init__()
        self.objects_supported = True

    def _serialize(self, obj: typing.Any) -> bytes:
        return pickle.dumps(obj)

    def _deserialize(self, data: bytes) -> typing.Any:
        return pickle.loads(data)


class JSONSerializer(Serializer):
    # TODO turn this into hybrid serializer that uses JSON for incoming and pickle for outgoing?
    # TODO how can we support exceptions?
    def _serialize(self, obj: typing.Any) -> bytes:
        try:
            return json.dumps(obj).encode("utf-8")
        except (TypeError, ValueError) as e:
            suggestion = "The default serializer doesn't support objects. If you need to pass objects, trust all the " \
                         "code that can call functions, and understand the risks of pickle, use app = lovage.Lovage(" \
                         "serializer=lovage.backends.PickleSerializer()) "
            raise LovageException(suggestion) from e  # TODO better exception type here

    def _deserialize(self, data: bytes) -> typing.Any:
        return json.loads(data.decode("utf-8"))


class Executor(object):
    def invoke(self, serializer: Serializer, func: types.FunctionType, packed_args):
        raise NotImplementedError()

    def invoke_async(self, serializer: Serializer, func: types.FunctionType, packed_args):
        raise NotImplementedError()

    def queue(self, serializer: Serializer, func: types.FunctionType, packed_args):
        raise NotImplementedError()

    def delay(self, serializer: Serializer, func: types.FunctionType, packed_args, timeout):
        raise NotImplementedError()


class Task(object):
    def __init__(self, func: types.FunctionType, executor: Executor, serializer: Serializer):
        self._func = func
        self._executor = executor
        self._serializer = serializer

    def __call__(self, *args, **kwargs):
        if is_in_cloud():
            raise LovageConfigurationError("Local backend used in cloud deployment. Make sure the right cloud backend "
                                           "is used in the code that runs in the cloud.")

        warnings.warn(f"{self._func.__name__} called directly. Use .invoke(), .invoke_async(), .queue() or .delay() "
                      f"to take advantage of Lovage.",
                      stacklevel=2)
        return self._func(*args, **kwargs)

    def call(self, *args, **kwargs):
        return self._func(*args, **kwargs)

    def invoke(self, *args, **kwargs):
        packed_args = self._serializer.pack_args(args, kwargs)
        packed_result = self._executor.invoke(self._serializer, self._func, packed_args)
        return self._serializer.unpack_result(packed_result)

    def invoke_async(self, *args, **kwargs):
        packed_args = self._serializer.pack_args(args, kwargs)
        self._executor.invoke_async(self._serializer, self._func, packed_args)

    def queue(self, *args, **kwargs):
        packed_args = self._serializer.pack_args(args, kwargs)
        self._executor.queue(self._serializer, self._func, packed_args)

    def delay(self, timeout, *args, **kwargs):
        packed_args = self._serializer.pack_args(args, kwargs)
        self._executor.delay(self._serializer, self._func, packed_args, timeout)


class Backend(object):
    def new_task(self, serializer: Serializer, func: types.FunctionType, options: typing.Mapping) -> Task:
        raise NotImplementedError()

    def deploy(self, *, requirements: typing.List[str], root: str, exclude=None):
        raise NotImplementedError()



# taken from https://github.com/tsileo/dirtools/
# updated for Python 3 and removed stuff we don't need

import logging
import os

from globster import Globster

log = logging.getLogger("dirtools")


# TODO abs=True args for .files(), .subdirs() ?


def load_patterns(exclude_file=".exclude"):
    """ Load patterns to exclude file from `exclude_file',
    and return a list of pattern.

    :type exclude_file: str
    :param exclude_file: File containing exclude patterns

    :rtype: list
    :return: List a patterns

    """
    return filter(None, open(exclude_file).read().split("\n"))


class Dir(object):
    """ Wrapper for dirtools arround a path.

    Try to load a .exclude file, ready to compute hashdir,


    :type directory: str
    :param directory: Root directory for initialization

    :type exclude_file: str
    :param exclude_file: File containing exclusion pattern,
        .exclude by default, you can also load .gitignore files.

    :type excludes: list
    :param excludes: List of additionals patterns for exclusion,
        by default: ['.git/', '.hg/', '.svn/']

    """

    def __init__(self, directory=".", exclude_file=".exclude",
                 excludes=['.git/', '.hg/', '.svn/'], includes=[]):
        if not os.path.isdir(directory):
            raise TypeError("Directory must be a directory.")
        self.directory = os.path.basename(directory)
        self.path = os.path.abspath(directory)
        self.parent = os.path.dirname(self.path)
        self.exclude_file = os.path.join(self.path, exclude_file)
        self.patterns = excludes
        if os.path.isfile(self.exclude_file):
            self.patterns.extend(load_patterns(self.exclude_file))
        self.globster = Globster(self.patterns)

    def is_excluded(self, path):
        """ Return True if `path' should be excluded
        given patterns in the `exclude_file'. """
        match = self.globster.match(self.relpath(path))
        if match:
            log.debug("{0} matched {1} for exclusion".format(path, match))
            return True
        return False

    def walk(self):
        """ Walk the directory like os.path
        (yields a 3-tuple (dirpath, dirnames, filenames)
        except it exclude all files/directories on the fly. """
        for root, dirs, files in os.walk(self.path, topdown=True):
            # TODO relative walk, recursive call if root excluder found???
            # root_excluder = get_root_excluder(root)
            ndirs = []
            # First we exclude directories
            for d in list(dirs):
                if self.is_excluded(os.path.join(root, d)):
                    dirs.remove(d)
                elif not os.path.islink(os.path.join(root, d)):
                    ndirs.append(d)

            nfiles = []
            for fpath in (os.path.join(root, f) for f in files):
                if not self.is_excluded(fpath) and not os.path.islink(fpath):
                    nfiles.append(os.path.relpath(fpath, root))

            yield root, ndirs, nfiles

    def relpath(self, path):
        """ Return a relative filepath to path from Dir path. """
        return os.path.relpath(path, start=self.path)



import functools
import os

import pkg_resources

import lovage.backends
import lovage.backends.base
import lovage.utils

try:
    __version__ = pkg_resources.get_distribution('lovage').version
except pkg_resources.DistributionNotFound:
    __version__ = '0.0.0'


class Lovage(object):
    def __init__(self,
                 backend: lovage.backends.base.Backend = lovage.backends.LocalBackend(),
                 serializer: lovage.backends.base.Serializer = lovage.backends.base.JSONSerializer()):
        self._backend = backend
        self._serializer = serializer

        self._functions = []

    def task(self, *args, **kwargs) -> lovage.backends.base.Task:
        def inner_create_task_cls(**opts):
            def _create_task_cls(func):
                task = self._backend.new_task(self._serializer, func, opts)
                return functools.update_wrapper(task, func)

            return _create_task_cls

        if len(args) == 1:
            if callable(args[0]):
                return inner_create_task_cls(**kwargs)(*args)
            raise TypeError('argument 1 to @task() must be a callable')
        if args:
            raise TypeError(f'@task() takes exactly 1 argument ({len(args) + len(kwargs)} given)')

        return inner_create_task_cls(**kwargs)

    def deploy(self, *, requirements="", root=os.getcwd(), exclude=None):
        if isinstance(requirements, str):
            requirements = [r.strip() for r in requirements.split("\n")]
        print(f"Deploying files...\n  root={root}\n  requirements={requirements}")
        self._backend.deploy(requirements=requirements, root=root, exclude=exclude)

    def is_local_backend(self):
        """
        Checks if this app is configured to run locally.
        :return: True if default local backend is being used
        """
        return isinstance(self._backend, lovage.backends.LocalBackend)

    @staticmethod
    def is_in_cloud():
        """
        Checks if this code is running in a deployed Lovage stack. Useful when you have to initialize global variables
        only in deployed code.
        :return: True if running in AWS/GCP/Azure/etc.
        """
        return lovage.utils.is_in_cloud()



import os


def is_in_cloud():
    """
    Checks if this code is running in a deployed Lovage stack. Useful when you have to initialize global variables
    only in deployed code.
    :return: True if running in AWS/GCP/Azure/etc.
    """
    return os.getenv("LOVAGE_IN_CLOUD", "0") == "1"



import hashlib
import json
import os
import platform
import shlex
import shutil
import subprocess
import traceback
import urllib.error
import urllib.request
import venv
import zipfile

import boto3

# TODO zip directly to s3 instead of /tmp first?

SUCCESS = "SUCCESS"
FAILED = "FAILED"


def cfn_response(event, context, status, physical_resource_id, data, reason=None):
    if reason:
        reason += "\n\n"
    else:
        reason = ""
    reason += f"See CloudWatch for details: {context.log_group_name} {context.log_stream_name}"

    response = json.dumps({
        "Status": status,
        "Reason": reason,
        "PhysicalResourceId": physical_resource_id,
        "StackId": event["StackId"],
        "RequestId": event["RequestId"],
        "LogicalResourceId": event["LogicalResourceId"],
        "Data": data or {}
    }).encode("utf-8")

    opener = urllib.request.build_opener(urllib.request.HTTPSHandler)
    request = urllib.request.Request(
        event["ResponseURL"],
        data=response,
        headers={
            "Content-Type": "",
            "Content-Length": f"{len(response)}"
        },
        method="PUT")
    opener.open(request)


def handler(event, context):
    pid = "BAD-PARAMETERS"

    try:
        requirements = _clean_requirements(event["ResourceProperties"]["Requirements"])
        hashed_data = requirements + " XX_VERSION_XX " + platform.python_version()
        rhash = hashlib.md5(hashed_data.encode("utf-8")).hexdigest()
        pid = f"req-{rhash}"
        key = f"requirements-{rhash}.zip"
        bucket = os.environ['BUCKET']

        if event["RequestType"] in ["Create", "Update"]:
            print(f"Installing on Python {platform.python_version()}: {requirements}...")

            shutil.rmtree("/tmp/venv", ignore_errors=True)
            shutil.rmtree("/tmp/python", ignore_errors=True)

            # we create a venv so package upgrades don't attempt read-only /var/runtime libraries
            venv.create("/tmp/venv", with_pip=True)
            cmd = f"/tmp/venv/bin/python -m pip --disable-pip-version-check install " \
                  f"-t /tmp/python --progress-bar off {requirements}"
            print(cmd)

            pip_result = subprocess.run(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True)
            if pip_result.returncode != 0:
                # response is limited to 4096 bytes total
                cfn_response(event, context, FAILED, pid, None,
                             f"pip failed [{pip_result.returncode}]:\n\n[...] {pip_result.stdout[-700:]}")
                return

            print(f"Building requirements package...")

            with zipfile.ZipFile("/tmp/python.zip", "w") as z:
                for root, folders, files in os.walk("/tmp/python"):
                    for f in files:
                        local_path = os.path.join(root, f)
                        zip_path = os.path.relpath(local_path, "/tmp")
                        z.write(local_path, zip_path, zipfile.ZIP_DEFLATED)

            print(f"Uploading to s3://{bucket}/{key}")

            boto3.client("s3").upload_file("/tmp/python.zip", bucket, key)

        elif event["RequestType"] == "Delete":
            print(f"Deleting s3://{bucket}/{key}")

            boto3.client('s3').delete_object(Bucket=bucket, Key=key)

        result = {"Bucket": bucket, "Key": key}
        cfn_response(event, context, SUCCESS, pid, result)
    except Exception as e:
        try:
            traceback.print_last()
        except ValueError:
            print("Caught exception but unable to print stack trace")
            print(e)
        cfn_response(event, context, FAILED, pid, None, str(e))


def _clean_requirements(requirements):
    result = ""
    for r in requirements:
        r = r.split("#")[0].strip()
        if r:
            result += f"{shlex.quote(r)} "
    return result



import threading
import time
import types
import typing
from concurrent.futures.thread import ThreadPoolExecutor

from . import base
from ..exceptions import LovageRemoteException


class LocalBackend(base.Backend):
    def __init__(self):
        self._executor = LocalExecutor()

    def new_task(self, serializer: base.Serializer, func: types.FunctionType, options: typing.Mapping) -> base.Task:
        return base.Task(func, self._executor, serializer)

    def deploy(self, *, requirements: typing.List[str], root: str, exclude=None):
        print("Nothing to deploy when running locally")


class LocalExecutor(base.Executor):
    def __init__(self):
        self._executor = ThreadPoolExecutor(max_workers=1)

    def invoke(self, serializer: base.Serializer, func: types.FunctionType, packed_args):
        return self._invoke(serializer, func, packed_args)

    def invoke_async(self, serializer: base.Serializer, func: types.FunctionType, packed_args):
        threading.Thread(target=self._invoke, args=(serializer, func, packed_args)).start()

    def queue(self, serializer: base.Serializer, func: types.FunctionType, packed_args):
        self._executor.submit(self._invoke, serializer, func, packed_args)

    def delay(self, serializer: base.Serializer, func: types.FunctionType, packed_args, timeout):
        def delayer():
            time.sleep(timeout)
            self._invoke(serializer, func, packed_args)

        threading.Thread(target=delayer).start()

    @staticmethod
    def _invoke(serializer: base.Serializer, func: types.FunctionType, packed_args):
        # TODO handle exceptions so we can test serializers
        try:
            unpacked_args, unpacked_kwargs = serializer.unpack_args(packed_args)
            result = func(*unpacked_args, **unpacked_kwargs)
            return serializer.pack_result(result)
        except Exception as e:
            # exception_handler(e) -- TODO AWS only for now
            if serializer.objects_supported:
                packed_e = serializer.pack_result(e)
                unpacked_e = serializer.unpack_result(packed_e)
                raise unpacked_e
            else:
                raise LovageRemoteException.from_exception_object(LovageRemoteException.exception_object(e))



import traceback


class LovageException(Exception):
    """
    Parent to all exceptions raised by Lovage.
    """
    pass


class LovageConfigurationError(LovageException):
    """
    Lovage configuration issue detected.
    """
    pass


class LovageDeploymentException(LovageException):
    """
    Deployment related errors.
    """
    pass


class LovageInternalException(LovageException):
    """
    Internal exception not related to user code.
    """
    pass


class LovageRemoteException(LovageException):
    """
    An exception describing an exception that was raised by the remote executed function.

    To get information about the original exception check out `exception` for the exception class name, `exception_fqn`
    for the fully qualified name, and `args` for the arguments the original exception had.
    """

    def __init__(self, exception, exception_fqn, exception_args, exception_str, stack_trace=None):
        # TODO copy stack trace
        super().__init__(*exception_args)
        self.exception = exception
        self.exception_fqn = exception_fqn
        self._str = exception_str

    def __str__(self):
        """
        :return: unmodified string representation of the remote exception
        """
        return self._str

    @staticmethod
    def exception_object(e: Exception):
        return {
            "exception": e.__class__.__name__,
            "exception_fqn": f"{e.__class__.__module__}.{e.__class__.__qualname__}",
            "exception_args": e.args,
            "exception_str": str(e),
        }

    @staticmethod
    def from_exception_object(o: dict):
        return LovageRemoteException(**o)



import os

import setuptools

# TODO lovage[runtime], lovage[aws], lovage[django]

version = os.getenv("GITHUB_REF")
if version and version.startswith("refs/tags/"):
    version = version.replace("refs/tags/", "")
else:
    version = "0.0.0"

setuptools.setup(
    name="lovage",
    version=version,
    description="Kind of like Celery, but simpler and with more Lambda",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    keywords="celery queue task lambda serverless",
    author="CloudSnorkel",
    author_email="amir@cloudsnorkel.com",
    url="https://github.com/CloudSnorkel/lovage",
    license="MIT",
    packages=setuptools.find_packages(),
    package_data={"lovage": ["backends/awslambda/helpers/*.py"]},
    include_package_data=True,
    zip_safe=True,
    python_requires=">=3.6",
    setup_requires=["wheel"],
    install_requires=[
        # TODO auto sync with pipenv?
        "boto3",
        "troposphere",
        "globster==0.1.0",
    ],
)



import contextlib
import hashlib
import pkgutil
import platform
import re
import typing

import boto3
import botocore.exceptions
import troposphere.awslambda
import troposphere.cloudformation
import troposphere.iam
import troposphere.logs
import troposphere.s3

from lovage.exceptions import LovageDeploymentException

REQUIREMENTS_LAYER_PACKAGER_CODE = pkgutil.get_data('lovage', 'backends/awslambda/helpers/packager.py').decode('utf-8')
CODE_DELETER_CODE = pkgutil.get_data('lovage', 'backends/awslambda/helpers/deleter.py').decode('utf-8')
assert REQUIREMENTS_LAYER_PACKAGER_CODE and CODE_DELETER_CODE


def _alphanumeric_name(name):
    return re.sub("[^a-zA-Z0-9]", "X", name)


class RequirementsLayerPackage(troposphere.cloudformation.AWSCustomObject):
    resource_type = "Custom::RequirementsLayerPackage"

    props = {
        'ServiceToken': (str, True),
    }


def _add_log_group(template: troposphere.Template, cf_name: str, name: str):
    return troposphere.logs.LogGroup(
        f"{cf_name}LogGroup",
        template,
        LogGroupName=troposphere.Sub(f"/aws/lambda/{name}"),
        RetentionInDays=30,  # TODO config
    )


def _get_python_runtime():
    # https://docs.aws.amazon.com/lambda/latest/dg/lambda-runtimes.html#w503aac27c25
    v = platform.python_version_tuple()
    vs = f"python{v[0]}.{v[1]}"
    if vs in ("python3.6", "python3.7", "python3.8"):
        return vs
    raise LovageDeploymentException(f"{vs} is not supported in AWS Lambda")


def _add_lambda(template: troposphere.Template, cf_name: str, name: str, policies: typing.List[troposphere.iam.Policy],
                **kwargs) -> troposphere.awslambda.Function:
    log_group = _add_log_group(template, cf_name, name)
    log_policy = troposphere.iam.Policy(
        PolicyName="Log",
        PolicyDocument={
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Action": [
                        "logs:CreateLogGroup",
                        "logs:CreateLogStream",
                        "logs:PutLogEvents",
                    ],
                    "Resource": [
                        troposphere.Sub(f"${{{log_group.title}.Arn}}"),
                        troposphere.Sub(f"${{{log_group.title}.Arn}}/*"),
                    ],
                },
            ]
        }
    )

    role = troposphere.iam.Role(f"{cf_name}Role", template)
    role.AssumeRolePolicyDocument = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Effect": "Allow",
                "Principal": {
                    "Service": [
                        troposphere.Sub("lambda.${AWS::URLSuffix}")
                    ]
                },
                "Action": [
                    "sts:AssumeRole"
                ]
            }
        ],
    }
    role.Policies = [log_policy] + policies

    func = troposphere.awslambda.Function(cf_name, template, **kwargs)
    if '${' in name:
        func.FunctionName = troposphere.Sub(name)
    else:
        func.FunctionName = name
    func.Runtime = _get_python_runtime()
    func.Role = role.get_att("Arn")
    func.DependsOn = [log_group.title]

    return func


def _add_str_lambda(template: troposphere.Template, name: str, code: str, policies: typing.List[troposphere.iam.Policy],
                    **kwargs):
    return _add_lambda(
        template, name, f"${{AWS::StackName}}-{name}", policies,
        Code=troposphere.awslambda.Code(ZipFile=code),
        Handler="index.handler",
        **kwargs
    )


def _add_codezip_lambda(template, cf_name, name, code, policies, **kwargs) -> troposphere.awslambda.Function:
    return _add_lambda(
        template, cf_name, name, policies,
        Code=troposphere.awslambda.Code(
            S3Bucket=troposphere.Sub(f"${{{code.title}.Bucket}}"),
            S3Key=troposphere.Sub(f"${{{code.title}.Key}}"),
        ),
        **kwargs
    )


class CodePackage(troposphere.cloudformation.AWSCustomObject):
    resource_type = "Custom::CodePackage"

    props = {
        'ServiceToken': (str, True),
    }


class TemplateFile(troposphere.cloudformation.AWSCustomObject):
    resource_type = "Custom::TemplateFile"

    props = {
        'ServiceToken': (str, True),
    }


def _stub_template():
    template = troposphere.Template()

    bucket = troposphere.s3.Bucket(
        "LovageBucket",
        template
    )

    code_deleter = _add_str_lambda(
        template,
        "LovageCodeDeleter",
        CODE_DELETER_CODE,
        [
            troposphere.iam.Policy(
                PolicyName="DeleteCode",
                PolicyDocument={
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Effect": "Allow",
                            "Action": [
                                "s3:DeleteObject",
                            ],
                            "Resource": [
                                troposphere.Sub("${LovageBucket.Arn}/code-*.zip"),
                                troposphere.Sub("${LovageBucket.Arn}/template.yml"),
                            ],
                        },
                    ]
                }
            )
        ],
        Environment=troposphere.awslambda.Environment(
            Variables={
                "BUCKET": bucket.ref(),
            }
        ),
        Description="Deletes code packages from bucket so bucket can be easily deleted",
    )

    TemplateFile(
        "LovageCfnTemplate",
        template,
        ServiceToken=code_deleter.get_att("Arn"),
        Key="template.yml",
    )

    return bucket, code_deleter, template


def generate_stub_template():
    _, _, template = _stub_template()
    return template.to_yaml(clean_up=True, long_form=True)


def generate_template(stack_name: str, bucket_name: str, code_key: str, requirements: typing.List[str],
                      functions: typing.Sequence[typing.Mapping],
                      resources: typing.Sequence[troposphere.BaseAWSObject],
                      env: typing.Dict[str, object],
                      policies: typing.Sequence):
    bucket, code_deleter, template = _stub_template()

    packager = _add_str_lambda(
        template,
        "LoaveRequirementsPackager",
        REQUIREMENTS_LAYER_PACKAGER_CODE,
        [
            troposphere.iam.Policy(
                PolicyName="SaveLayer",
                PolicyDocument={
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Effect": "Allow",
                            "Action": [
                                "s3:PutObject",
                                "s3:DeleteObject",
                            ],
                            "Resource": troposphere.Sub("${LovageBucket.Arn}/requirements-*.zip"),
                        },
                    ]
                }
            )
        ],
        Environment=troposphere.awslambda.Environment(
            Variables={
                "BUCKET": bucket.ref()
            }
        ),
        Timeout=15 * 60,
        MemorySize=1024,
        Description="Downloads Python requirements, zips them, and uploads to a bucket to be used by Lambda layer",
    )

    package = RequirementsLayerPackage(
        "LovageRequirementsPackage",
        template,
        ServiceToken=packager.get_att("Arn"),
        Requirements=requirements,
        # we need to rebuild requirements.zip if python version changes because it might install different libraries
        PythonVersion=_get_python_runtime(),
    )

    layer = troposphere.awslambda.LayerVersion(
        f"{_alphanumeric_name(stack_name)}RequirementsLayer",
        template,
        Content=troposphere.awslambda.Content(
            S3Bucket=troposphere.Sub(f"${{{package.title}.Bucket}}"),
            S3Key=troposphere.Sub(f"${{{package.title}.Key}}"),
        )
    )

    code = CodePackage(
        "LovageCodePackage",
        template,
        ServiceToken=code_deleter.get_att("Arn"),
        Key=code_key,
    )

    for f in functions:
        lf = _add_codezip_lambda(
            template,
            f["CfName"],
            f["Name"],
            code,
            [
                troposphere.iam.Policy(
                    PolicyName=f"Custom{i}",
                    PolicyDocument=p)
                for i, p in enumerate(policies + f["Policies"])
            ],
            Layers=[layer.ref()],
            Handler=f["Handler"],
            **f["Kwargs"],
        )

        lf.Environment = troposphere.awslambda.Environment(Variables=env)

    for r in resources:
        template.add_resource(r)

    return template.to_yaml(clean_up=True, long_form=True)


def _stack_exists(cf, name):
    try:
        cf.describe_stacks(StackName=name)
        return True
    except botocore.exceptions.ClientError as e:
        if e.response["Error"]["Code"] == "ValidationError":
            return False
        raise


@contextlib.contextmanager
def _code_uploader(session, bucket, code_bytes):
    print("Uploading code...")

    code_hash = hashlib.md5(code_bytes).hexdigest()
    code_key = f"code-{code_hash}.zip"

    delete_on_failure = False

    s3 = session.client("s3")

    try:
        s3.head_object(Bucket=bucket, Key=code_key)
        print("Code already uploaded")
    except botocore.exceptions.ClientError as e:
        if e.response["Error"]["Code"] == "404":
            s3.put_object(Body=code_bytes, Bucket=bucket, Key=code_key, ContentType="application/zip")
            # only delete on failure if we uploaded the code and it's not the old code
            delete_on_failure = True
        else:
            raise

    try:

        yield code_key

    except Exception:
        if delete_on_failure:
            try:
                s3.delete_object(Bucket=bucket, Key=code_key)
                print("Code already uploaded")
            except botocore.exceptions.ClientError as e:
                print("Error while deleting code package", e)

        raise


def wait_and_log(cf, waiter, stack_name):
    try:
        cf.get_waiter(waiter).wait(StackName=stack_name)
        return
    except botocore.exceptions.WaiterError as e:
        # TODO check that e is stack failure and not something else
        reason = f"Stack {stack_name} failed to deploy due to:"
        for events_page in cf.get_paginator("describe_stack_events").paginate(StackName=stack_name):
            for event in events_page["StackEvents"]:
                if event["LogicalResourceId"] == stack_name and event.get("ResourceStatusReason") == "User Initiated" \
                        and event["ResourceStatus"] in ["UPDATE_IN_PROGRESS", "CREATE_IN_PROGRESS"]:
                    raise LovageDeploymentException(reason) from None
                if event["ResourceStatus"] in ["CREATE_FAILED", "DELETE_FAILED", "UPDATE_FAILED"]:
                    if not event["ResourceStatusReason"]:
                        # empty error
                        continue
                    if event["ResourceStatusReason"] in ["Resource creation cancelled", "Resource update cancelled"]:
                        # this "error" doesn't help debugging
                        continue
                    reason += "\n  %(LogicalResourceId)s | %(ResourceStatus)s | %(ResourceStatusReason)s" % event
                    # TODO if it's a custom resource error, get the custom resource log too
        raise LovageDeploymentException(reason) from None


def deploy(session: boto3.Session, stack_name: str, code_bytes: bytes, requirements: typing.List[str],
           functions: typing.Sequence[typing.Mapping],
           resources: typing.Sequence[troposphere.BaseAWSObject],
           env: typing.Dict[str, object],
           policies: typing.Sequence):
    cf = session.client("cloudformation")

    if not _stack_exists(cf, stack_name):
        print("Creating stub stack...")
        cf.create_stack(
            StackName=stack_name,
            TemplateBody=generate_stub_template(),
            Tags=[
                {
                    "Key": "Lovage",
                    "Value": "true",  # TODO version?
                },
            ],
            Capabilities=["CAPABILITY_IAM"],
        )

        wait_and_log(cf, "stack_create_complete", stack_name)

    for t in cf.describe_stacks(StackName=stack_name)["Stacks"][0]["Tags"]:
        if t["Key"] == "Lovage":
            break
    else:
        raise ValueError(f"Stack `{stack_name}` already exists. Use a different deployment name.")

    bucket = cf.describe_stack_resource(
        StackName=stack_name, LogicalResourceId="LovageBucket")["StackResourceDetail"]["PhysicalResourceId"]

    try:
        with _code_uploader(session, bucket, code_bytes) as code_key:
            print("Uploading template...")
            tmpl = generate_template(stack_name, bucket, code_key, requirements, functions, resources, env, policies)
            session.client("s3").put_object(Body=tmpl, Bucket=bucket, Key="template.yml", ContentType="text/yaml")

            print("Updating stack...")
            cf.update_stack(
                StackName=stack_name,
                TemplateURL=f"https://s3.amazonaws.com/{bucket}/template.yml",
                Capabilities=["CAPABILITY_IAM"],
                Parameters=[],
            )

            wait_and_log(cf, "stack_update_complete", stack_name)
    except botocore.exceptions.ClientError as e:
        if e.response['Error']['Code'] == 'ValidationError' \
                and e.response['Error']['Message'] == 'No updates are to be performed.':
            print("Stack already up-to-date")
            return
        raise
